@@ -18,13 +18,15 @@
 # the GNU General Public License along with this program.  If not,
 # see <http://www.lsstcorp.org/LegalNotices/>.
 
+import contextlib
 import json
 import logging
 import os
 import sys
+import warnings
 from collections import OrderedDict
 from importlib.metadata import PackageNotFoundError, version
-from typing import IO, Any
+from typing import ContextManager, Optional, TextIO, cast
 
 import arrow
 import click
@@ -44,45 +46,131 @@ from .tplan import build_tpr_model, render_report
 from .vcd import build_vcd_dict, summary
 from .ve_baseline import do_ve_model
 
-__version__: str
+# Silence BeautifulSoup warning about providing Unicode markup together
+# with from_encoding
+warnings.filterwarnings(
+    "ignore",
+    message=(
+        r"You provided Unicode markup but also provided a value for "
+        r"from_encoding\. Your from_encoding will be ignored\."
+    ),
+)
 
 try:
     __version__ = version(__name__)
 except PackageNotFoundError:
-    # package is not installed
     __version__ = "0.0.0"
 
-# Global variables
-OUTPUT_FORMAT: str = "latex"
+OUTPUT_FORMAT = "latex"
 
 if "ZEPHYR_TOKEN" in os.environ:
     Config.ZEPHYR_TOKEN = os.environ["ZEPHYR_TOKEN"]
-
 if "JIRA_PASSWORD" in os.environ:
     Config.AUTH = (os.environ["JIRA_USER"], os.environ["JIRA_PASSWORD"])
 
-# Suppress BSoup/logging warnings
 logging.getLogger("root").setLevel(logging.ERROR)
 logging.getLogger().setLevel(logging.ERROR)
 logging.getLogger("bs4.dammit").setLevel(logging.ERROR)
 
 
+def _as_output_format(text: str) -> str:
+    """Return converted text when template language differs from
+    output format.
+    """
+    if Config.TEMPLATE_LANGUAGE != OUTPUT_FORMAT:
+        setattr(Config.DOC, Config.TEMPLATE_LANGUAGE, text.encode("utf-8"))
+        return getattr(Config.DOC, OUTPUT_FORMAT).decode("utf-8")
+    return text
+
+
+def _env() -> Environment:
+    return Environment(
+        loader=ChoiceLoader(
+            [
+                FileSystemLoader(Config.TEMPLATE_DIRECTORY),
+                PackageLoader("docsteady", "templates"),
+            ]
+        ),
+        lstrip_blocks=True,
+        trim_blocks=True,
+        autoescape=False,
+    )
+
+
+def _load_template(env: Environment, target: str) -> Optional[Template]:
+    path = f"{target}.{Config.TEMPLATE_LANGUAGE}.jinja2"
+    try:
+        return env.get_template(path)
+    except TemplateNotFound:
+        click.echo(f"No Template Found: {path}", err=True)
+        return None
+
+
+def _write_output(text: str, path: Optional[str]) -> None:
+    if Config.TEMPLATE_LANGUAGE != OUTPUT_FORMAT:
+        setattr(Config.DOC, Config.TEMPLATE_LANGUAGE, text.encode("utf-8"))
+        text = getattr(Config.DOC, OUTPUT_FORMAT).decode("utf-8")
+    if path:
+        with open(path, "w", encoding="utf-8") as fp:
+            fp.write(text)
+    else:
+        sys.stdout.write(text)
+
+
+def _try_appendix_template(
+    env: Environment, target: str
+) -> Optional[Template]:
+    try:
+        return env.get_template(
+            f"{target}-appendix.{Config.TEMPLATE_LANGUAGE}.jinja2"
+        )
+    except TemplateNotFound:
+        return None
+
+
+def _get_appendix_path(path: Optional[str]) -> ContextManager[TextIO]:
+    """Return a context manager that yields a writable TextIO.
+
+    If `path` is None or doesn't look like a filename with an
+    extension, yield sys.stdout via contextlib.nullcontext. Otherwise
+    open the appendix file for writing and yield that file object.
+    """
+    if not path:
+        return cast(ContextManager[TextIO], contextlib.nullcontext(sys.stdout))
+    parts = path.split(".")
+    if len(parts) < 2:
+        return cast(ContextManager[TextIO], contextlib.nullcontext(sys.stdout))
+    appendix = ".".join(parts[:-1] + ["appendix", parts[-1]])
+    return open(appendix, "w", encoding="utf-8")
+
+
+def _metadata() -> dict:
+    return {
+        "created_on": arrow.now(),
+        "docsteady_version": __version__,
+        "project": "LVV",
+    }
+
+
 @click.group()
 @click.option(
     "--namespace",
     default="dm",
-    help="Project namespace (dm, ts, example, etc..). " 'Defaults to "dm".',
+    help=("Project namespace (dm, ts, example, etc..). " "Defaults to dm."),
 )
 @click.option(
     "--template-format",
+    "template_format",
     default="latex",
-    help="Template language (latex, html). " 'Defaults to "latex".',
+    help=("Template language (latex, html). Defaults to latex."),
 )
 @click.option(
     "--load-from",
+    "load_from",
     default=os.path.curdir,
-    help="Path to search for templates in. "
-    "Defaults to the working directory",
+    help=(
+        "Path to search for templates in. Defaults to the working " "directory"
+    ),
 )
 @click.option(
     "--token",
@@ -96,14 +184,14 @@ logging.getLogger("bs4.dammit").setLevel(logging.ERROR)
     prompt="Jira User Name for Jira API",
     hide_input=True,
     envvar="JIRA_USER",
-    help="Jira cloud user - an email address ",
+    help="Jira cloud user - an email address",
 )
 @click.option(
     "--password",
     prompt="Jira password (or Token)  for Jira API",
     hide_input=True,
     envvar="JIRA_PASSWORD",
-    help="Jira cloud  password - usually an API token  ",
+    help="Jira cloud password - usually an API token",
 )
 @click.version_option(__version__)
 def cli(
@@ -114,9 +202,6 @@ def cli(
     username: str,
     password: str,
 ) -> None:
-    """Docsteady generates documents from Jira with the
-    Test Management for Jira (TM4J) plugin.
-    """
     Config.MODE_PREFIX = f"{namespace.lower()}-" if namespace else ""
     Config.NAMESPACE = namespace
     Config.TEMPLATE_LANGUAGE = template_format
@@ -128,43 +213,23 @@ def cli(
 @cli.command("generate-spec")
 @click.option(
     "--format",
+    "format_",
     default="latex",
     help="Pandoc output format (see pandoc for options)",
 )
 @click.argument("folder")
 @click.argument("path", required=False, type=click.Path())
-def generate_spec(
-    format: str, username: str, password: str, folder: str, path: str
-) -> None:
-    """Read in tests from TM4J plugin where FOLDER
-    is the TM4J Test Case Folder. If specified, PATH is the resulting
-    output.
-
-    If PATH is specified, docsteady will examine the output filename
-    and attempt to write an appendix to a similar file.
-    For example, if the output is jira_docugen.tex, the output
-    will also print out a jira_docugen.appendix.tex file if a
-    template for the appendix is found. Otherwise, it will print
-    to standard out.
-    """
+def generate_spec(format_: str, folder: str, path: Optional[str]) -> None:
     global OUTPUT_FORMAT
-    OUTPUT_FORMAT = format
-    Config.AUTH = (username, password)
+    OUTPUT_FORMAT = format_
     target = "spec"
-    # Commented this line out since it seems to never be used.
-    # Config.output = TemporaryFile(mode="r+")
 
-    # Build model
     try:
         testcases, requirements, tcs_dict = build_spec_model(folder)
-    except Exception as e:
-        print("Error in building model")
-        print(e)
-        raise e
+    except Exception as exc:
+        click.echo("Error in building model", err=True)
+        raise exc
 
-    file = open(path, "w") if path else sys.stdout
-
-    # Sort the dictionary
     requirements_to_testcases = OrderedDict(
         sorted(
             Config.REQUIREMENTS_TO_TESTCASES.items(),
@@ -172,32 +237,17 @@ def generate_spec(
         )
     )
 
-    env = Environment(
-        loader=ChoiceLoader(
-            [
-                FileSystemLoader(Config.TEMPLATE_DIRECTORY),
-                PackageLoader("docsteady", "templates"),
-            ]
-        ),
-        lstrip_blocks=True,
-        trim_blocks=True,
-        autoescape=False,  # Was None.
-    )
-
-    try:
-        template_path = f"{target}.{Config.TEMPLATE_LANGUAGE}.jinja2"
-        template = env.get_template(template_path)
-    except TemplateNotFound:
-        click.echo(f"No Template Found: {template_path}", err=True)
+    env = _env()
+    template = _load_template(env, target)
+    if not template:
         sys.exit(1)
 
     libtestcases = sorted(
-        Config.CACHED_LIBTESTCASES.values(), key=lambda testc: testc["keyid"]
+        Config.CACHED_LIBTESTCASES.values(), key=lambda t: t["keyid"]
     )
-
     metadata = _metadata()
-    metadata["folder"] = folder
-    metadata["template"] = template.filename
+    metadata.update(folder=folder, template=template.filename)
+
     text = template.render(
         metadata=metadata,
         deprecated=testcases["deprecated"],
@@ -208,233 +258,186 @@ def generate_spec(
         tc_status_list=Config.TESTCASE_STATUS_LIST,
         testcases_map=Config.CACHED_TESTCASES,
     )
+    _write_output(text, path)
 
-    print(_as_output_format(text), file=file)
-
-    # Will exit if it can't find a template
-    appendix_template = _try_appendix_template(target, env)
-    if not appendix_template:
+    appendix = _try_appendix_template(env, target)
+    if not appendix:
         click.echo("No Appendix Template Found, skipping...", err=True)
-        sys.exit(0)
-    metadata["template"] = appendix_template.filename
-    appendix_file = _get_appendix_output(path)
-    appendix_text = appendix_template.render(
-        metadata=metadata,
-        testcases=testcases,
-        requirements_to_testcases=requirements_to_testcases,
-        requirements_map=requirements,
-        testcases_map=Config.CACHED_TESTCASES,
-    )
-    print(_as_output_format(appendix_text), file=appendix_file)
+        return
+
+    metadata["template"] = appendix.filename
+    with _get_appendix_path(path) as ap:
+        appendix_text = appendix.render(
+            metadata=metadata,
+            testcases=testcases,
+            requirements_to_testcases=requirements_to_testcases,
+            requirements_map=requirements,
+            testcases_map=Config.CACHED_TESTCASES,
+        )
+        ap.write(_as_output_format(appendix_text) if False else appendix_text)
 
 
 @cli.command("generate-tpr")
 @click.option(
     "--excludenoexec",
+    "excludenoexec",
+    is_flag=True,
     default=False,
-    help="Ignore the test execution steps not executed/with no comment"
-    'Defaults to "False".',
+    help=("Ignore the test execution steps not executed/with no " "comment"),
 )
 @click.option(
     "--includeall",
+    "includeall",
+    is_flag=True,
     default=False,
-    help="Ignore the include in report field for executions and include all"
-    'Defaults to "False".',
+    help=(
+        "Ignore the include in report field for executions and " "include all"
+    ),
 )
 @click.option(
     "--format",
+    "format_",
     default="latex",
     help="Pandoc output format (see pandoc for options)",
 )
 @click.option(
     "--trace",
+    "trace",
+    is_flag=True,
     default=False,
-    help="If true, traceability table will be added in appendix",
+    help=("If true, traceability table will be added in appendix"),
 )
 @click.option(
     "--dump",
+    "dump",
     default=False,
-    help="If true, dump json before rendering tex, "
-    "if the file exists use it next time instead of hitting server",
+    help=(
+        "If true, dump json before rendering tex, if the file "
+        "exists use it next time instead of hitting server"
+    ),
 )
 @click.argument("plan")
 @click.argument("path", required=False, type=click.Path())
 def generate_report(
-    format: str,
-    trace: str,
+    format_: str,
+    trace: bool,
     plan: str,
-    path: str,
+    path: Optional[str],
     includeall: bool,
     excludenoexec: bool,
     dump: bool,
 ) -> None:
-    """Read in a Test Plan and related cycles from TM4J.
-    If specified, PATH is the resulting output.
-    """
     global OUTPUT_FORMAT
-    OUTPUT_FORMAT = format
+    OUTPUT_FORMAT = format_
     Config.INCLUDE_ALL_EXECS = includeall
     target = "tpr"
 
-    if Config.NAMESPACE.upper() not in Config.COMPONENTS.keys():
-        print(f"Wrong input component {Config.NAMESPACE}")
-        exit()
-
-    # Commented this line out since it seems to never be used.
-    # Config.output = TemporaryFile(mode="r+")
+    if Config.NAMESPACE.upper() not in Config.COMPONENTS:
+        click.echo(f"Wrong input component {Config.NAMESPACE}", err=True)
+        sys.exit(1)
 
     fname = "tpr_model.json"
     if dump and os.path.isfile(fname):
-        with open(fname, "r") as fp:
+        with open(fname, "r", encoding="utf-8") as fp:
             plan_dict = json.load(fp)
     else:
         plan_dict = build_tpr_model(plan)
-        with open(fname, "w") as fp:
-            json.dump(plan_dict, fp)
+        with open(fname, "w", encoding="utf-8") as fp:
+            fp.write(json.dumps(plan_dict))
 
     metadata = _metadata()
     metadata["namespace"] = Config.NAMESPACE
     metadata["component_long_name"] = Config.COMPONENTS[
         Config.NAMESPACE.upper()
     ]
-    logging.log(logging.INFO, f"Rendering  {path}")
-    env = render_report(
+
+    render_report(
         excludenoexec, metadata, target, plan_dict, OUTPUT_FORMAT, path
     )
 
     # output the plan - TR without results
     target = "tpnoresult"
-    path = path.replace(".tex", "-plan.tex")
-    logging.log(logging.INFO, f"Rendering  {path}")
-    env = render_report(
-        excludenoexec, metadata, target, plan_dict, OUTPUT_FORMAT, path
+    path_plan = (path or "output.tex").replace(".tex", "-plan.tex")
+    render_report(
+        excludenoexec, metadata, target, plan_dict, OUTPUT_FORMAT, path_plan
     )
+
     if trace:
-        # Will exit if it can't find a template
-        appendix_template = _try_appendix_template(target, env)
-        if not appendix_template:
-            click.echo("No Appendix Template Found, skipping...", err=True)
-            sys.exit(0)
-        metadata["template"] = appendix_template.filename
-        appendix_file = _get_appendix_output(path)
-        appendix_text = appendix_template.render(
-            metadata=metadata, testcases_map=plan_dict["testcases_map"]
-        )
-        print(_as_output_format(appendix_text), file=appendix_file)
+        env = _env()
+        appendix = _try_appendix_template(env, target)
+        if appendix:
+            metadata["template"] = appendix.filename
+            with _get_appendix_path(path_plan) as ap:
+                appendix_text = appendix.render(
+                    metadata=metadata,
+                    testcases_map=plan_dict["testcases_map"],
+                )
+                ap.write(appendix_text)
 
     if Config.exeuction_errored:
         raise SystemError("Content Problem, please check.")
 
 
-def _try_appendix_template(target: str, env: Environment) -> Template | None:
-    # Now appendix
-    appendix_template_path = (
-        f"{target}-appendix.{Config.TEMPLATE_LANGUAGE}.jinja2"
-    )
-
-    try:
-        return env.get_template(appendix_template_path)
-    except TemplateNotFound:
-        return None
-
-
-def _get_appendix_output(path: str) -> IO:
-    appendix_path = None
-    if path:
-        parts = path.split(".")
-        extension = parts[-1]
-        path_parts = parts[:-1] + ["appendix", extension]
-        appendix_path = ".".join(path_parts)
-    return open(appendix_path, "w") if appendix_path else sys.stdout
-
-
-def _as_output_format(text: str) -> str:
-    if Config.TEMPLATE_LANGUAGE != OUTPUT_FORMAT:
-        setattr(Config.DOC, Config.TEMPLATE_LANGUAGE, text.encode("utf-8"))
-        text = getattr(Config.DOC, OUTPUT_FORMAT).decode("utf-8")
-    return text
-
-
-def _metadata() -> dict:
-    return dict(
-        created_on=arrow.now(), docsteady_version=__version__, project="LVV"
-    )
-
-
 @cli.command("generate-vcd")
 @click.option(
     "--format",
+    "format_",
     default="latex",
     help="Pandoc output format (see pandoc for options)",
 )
 @click.option(
     "--spec",
-    required=False,
+    "spec",
+    is_flag=True,
     default=False,
-    help="Req|Test specifications to print out test case prioritization",
+    help=("Req|Test specifications to print out test case " "prioritization"),
 )
 @click.option(
     "--subcomponent",
+    "subcomponent",
     required=False,
-    help="Extract Verification Elements only "
-    "for the specified subcomponent",
+    help=(
+        "Extract Verification Elements only for the specified " "subcomponent"
+    ),
 )
 @click.option(
     "--dump",
     default=False,
-    help="If true, dump json before rendering tex, "
-    "if the file exists use it next time instead of hitting server",
+    help=(
+        "If true, dump json before rendering tex, if the file "
+        "exists use it next time instead of hitting server"
+    ),
 )
 @click.argument("path", required=False, type=click.Path())
 def generate_vcd(
-    format: str, spec: str, subcomponent: str, path: str, dump: bool
+    format_: str,
+    spec: bool,
+    subcomponent: Optional[str],
+    path: Optional[str],
+    dump: bool,
 ) -> None:
-    """Given a specific namespace, correspoding to a Jira Component
-    or Rubin Subsystem, it build the VCD. By default build the DM VCD.
-    If specified, PATH is the resulting output.
-    """
     global OUTPUT_FORMAT
-    OUTPUT_FORMAT = format
+    OUTPUT_FORMAT = format_
     target = "vcd"
-
     component = Config.NAMESPACE.upper()
-
-    print("Building VCD using Rest API access (VE extraction).")
-    if not subcomponent:
-        subcomponent = ""
+    subcomponent = subcomponent or ""
 
     if dump:
-        with open("VEmodel.json", "r") as fp:
+        with open("VEmodel.json", "r", encoding="utf-8") as fp:
             ve_model = json.load(fp)
     else:
         ve_model = do_ve_model(component, subcomponent)
+
     vcd_dict = build_vcd_dict(ve_model, usedump=dump)
-    sum_dict: list[dict | Any] = summary(vcd_dict)
+    sum_dict = summary(vcd_dict)
 
-    file = open(path, "w") if path else sys.stdout
-
-    env = Environment(
-        loader=ChoiceLoader(
-            [
-                FileSystemLoader(Config.TEMPLATE_DIRECTORY),
-                PackageLoader("docsteady", "templates"),
-            ]
-        ),
-        lstrip_blocks=True,
-        trim_blocks=True,
-        autoescape=False,
-    )
-
-    try:
-        template_path = f"{target}.{Config.TEMPLATE_LANGUAGE}.jinja2"
-        template = env.get_template(template_path)
-    except TemplateNotFound:
-        click.echo(f"No Template Found: {template_path}", err=True)
+    env = _env()
+    template = _load_template(env, target)
+    if not template:
         sys.exit(1)
 
     metadata = _metadata()
-    metadata["component"] = component
-    metadata["template"] = template.filename
+    metadata.update(component=component, template=template.filename)
     text = template.render(
         metadata=metadata,
         coverage=Config.req_coverage,
@@ -443,125 +446,101 @@ def generate_vcd(
         spec_to_reqs=Config.REQ_PER_DOC,
         vcd_dict=vcd_dict,
     )
-
-    print(_as_output_format(text), file=file)
-
-
-if __name__ == "__main__":
-    cli()
+    _write_output(text, path)
 
 
 @cli.command("baseline-ve")
 @click.option(
     "--format",
+    "format_",
     default="latex",
     help="Pandoc output format (see pandoc for options)",
 )
 @click.option(
     "--details",
+    "details",
+    is_flag=True,
     default=False,
-    help="If true, an extra detailed report will be produced",
+    help=("If true, an extra detailed report will be produced"),
 )
 @click.option(
     "--subcomponent",
+    "subcomponent",
     required=False,
-    help="Extract Verification Elements only "
-    "for the specified subcomponent",
+    help=(
+        "Extract Verification Elements only for the specified " "subcomponent"
+    ),
 )
 @click.option(
     "--dump",
     default=False,
-    help="If true, dump json before rendering tex, "
-    "if the file exists use it next time instead of hitting server",
+    help=(
+        "If true, dump json before rendering tex, if the file "
+        "exists use it next time instead of hitting server"
+    ),
 )
 @click.argument("path", required=False, type=click.Path())
 def baseline_ve(
-    format: str,
-    details: str,
+    format_: str,
+    details: bool,
     dump: bool,
-    subcomponent: str,
-    path: str,
+    subcomponent: Optional[str],
+    path: Optional[str],
 ) -> None:
-    """Given a specific subsystem (component), and subcomponent,
-    a document is generated including all corresponding Verification Elements
-    and related Test Cases. This is not a Verification Control Document:
-    no Test Result information is provided
-    """
     global OUTPUT_FORMAT
-    OUTPUT_FORMAT = format
+    OUTPUT_FORMAT = format_
     target = "ve"
     jfile = f"baseline_{target}.json"
-
     component = Config.NAMESPACE.upper()
+    subcomponent = subcomponent or ""
 
-    if not subcomponent:
-        subcomponent = ""
-
-    # it takes long time to get the data and sometimes it fails on render
-    # this means not going bakc to repeat all the jira calls saves an hour.
     if dump and os.path.exists(jfile):
-        with open(jfile, "r") as fp:
+        with open(jfile, "r", encoding="utf-8") as fp:
             ve_model = json.load(fp)
     else:
         ve_model = do_ve_model(component, subcomponent)
-        with open(jfile, "w") as f:
-            json.dump(ve_model, f)
+        with open(jfile, "w", encoding="utf-8") as fp:
+            fp.write(json.dumps(ve_model))
 
-    file = open(path, "w") if path else sys.stdout
-
-    env = Environment(
-        loader=ChoiceLoader(
-            [
-                FileSystemLoader(Config.TEMPLATE_DIRECTORY),
-                PackageLoader("docsteady", "templates"),
-            ]
-        ),
-        lstrip_blocks=True,
-        trim_blocks=True,
-        autoescape=False,
-    )
-
-    try:
-        template_path = f"{target}.{Config.TEMPLATE_LANGUAGE}.jinja2"
-        template = env.get_template(template_path)
-    except TemplateNotFound:
-        click.echo(f"No Template Found: {template_path}", err=True)
+    env = _env()
+    template = _load_template(env, target)
+    if not template:
         sys.exit(1)
 
     metadata = _metadata()
-    metadata["component"] = component
-    metadata["subcomponent"] = subcomponent
-    metadata["template"] = template.filename
+    metadata.update(
+        component=component,
+        subcomponent=subcomponent,
+        template=template.filename,
+    )
     text = template.render(
         metadata=metadata,
         velements=ve_model,
         reqs=Config.CACHED_REQS_FOR_VES,
         test_cases=Config.CACHED_TESTCASES,
     )
+    _write_output(text, path)
 
-    print(_as_output_format(text), file=file)
-    file.close()
-
-    # Writing detailed VE document
     if details:
-        details_file_name = "ve_details.tex"
-        details_file = open(details_file_name, "w")
         try:
-            template_path = (
+            template_details = env.get_template(
                 f"{target}-details.{Config.TEMPLATE_LANGUAGE}.jinja2"
             )
-            template_details = env.get_template(template_path)
+            details_text = template_details.render(
+                metadata=metadata,
+                velements=ve_model,
+                reqs=Config.CACHED_REQS_FOR_VES,
+                test_cases=Config.CACHED_TESTCASES,
+            )
+            with open("ve_details.tex", "w", encoding="utf-8") as df:
+                df.write(details_text)
         except TemplateNotFound:
             click.echo(
-                f"No Detailed template found: {template_path}", err=True
+                f"No Detailed template found: {target}-details."
+                f"{Config.TEMPLATE_LANGUAGE}.jinja2",
+                err=True,
             )
 
-        text_details = template_details.render(
-            metadata=metadata,
-            velements=ve_model,
-            reqs=Config.CACHED_REQS_FOR_VES,
-            test_cases=Config.CACHED_TESTCASES,
-        )
 
-        print(_as_output_format(text_details), file=details_file)
-        details_file.close()
+if __name__ == "__main__":
+    cli()
